// Copyright (C) 2019-2020  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// wcfs_misc.{h,cpp} provide miscellaneous utilities for other wcfs_* files.

#ifndef _NXD_WCFS_MISC_H_
#define _NXD_WCFS_MISC_H_

// XXX hack: C++ does not have __builtin_types_compatible_p, but CCAN configure
// thinks it does because CCAN is configired via C, not C++.
#include <config.h>
#undef  HAVE_BUILTIN_TYPES_COMPATIBLE_P
#define HAVE_BUILTIN_TYPES_COMPATIBLE_P 0
#include <ccan/array_size/array_size.h>

#include <stddef.h>
#include <stdint.h>

#include <golang/libgolang.h>
using namespace golang;

#include <string>
using std::string;

#include <utility>
using std::pair;
using std::make_pair;

#include <tuple>
using std::tuple;
using std::make_tuple;
using std::tie;

#include <vector>
using std::vector;


// nil is synonym for nullptr and NULL.	XXX -> pygolang
const nullptr_t nil = nullptr;


// io::     XXX -> pygolang
namespace io {

extern const global<error> EOF_;
extern const global<error> ErrUnexpectedEOF;


}   // io::


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// os::
namespace os {

// os::File mimics os.File from Go.
// its operations return error with full file context.
typedef refptr<class _File> File;
class _File : public object {
    int    _fd;
    string _path;

    // don't new - create via open
private:
    _File();
    ~_File();
    friend tuple<File, error> open(const string &path, int flags, mode_t mode);
public:
    void decref();

public:
    int     fd()    const;
    string  name()  const;
    error   close();

    // read implements io.Reader from Go: it reads into buf up-to count bytes.
    //
    // XXX buf -> slice<byte> ?
    tuple<int, error> read(void *buf, size_t count);

    // write implements io.Writer from Go: it writes all data from buf.
    //
    // NOTE write behaves like io.Writer in Go - it tries to write as much
    // bytes as requested, and if it could write only less - it returns error.
    // XXX buf -> slice<byte> ?
    tuple<int, error> write(const void *buf, size_t count);

    error   stat(struct stat *st);

private:
    error _errno(const char *op);
};

// open opens file @path.
tuple<File, error> open(const string &path, int flags = O_RDONLY,
        mode_t mode = S_IRUSR | S_IWUSR | S_IXUSR |
                      S_IRGRP | S_IWGRP | S_IXGRP |
                      S_IROTH | S_IWOTH | S_IXOTH);

}   // os::

// mm::
namespace mm {
    tuple<uint8_t*, error> map(int prot, int flags, os::File f, off_t offset, size_t size);
    error map_into(void *addr, size_t size, int prot, int flags, os::File f, off_t offset);
    error unmap(void *addr, size_t size);

}   // mm::



// ---- misc ----


// xstrconv::
namespace xstrconv {

tuple<uint64_t, error> parseHex64(const string& s);
tuple<int64_t,  error> parseInt(const string& s);
tuple<uint64_t, error> parseUint(const string& s);

}   // xstrconv::


// zodb::
namespace zodb {

typedef uint64_t Tid;
typedef uint64_t Oid;

}   // zodb::


// xerr::
namespace xerr {

// xerr::Contextf mimics xerr.Contextf from Go.
//
// Usage is a bit different(*) compared to Go:
//
//  func dosSomething(arg) {
//      xerr.Contextf E("doing something %s", v(arg));
//      ...
//      return E(err);
//  }
//
// (*) because C++ does not allow to modify returned value on the fly.
class Contextf {
    string errctx;

public:
    Contextf(const char *format, ...);
    error operator() (error) const;
};

}


// wcfs::
namespace wcfs {

// TidHead is invalid Tid which is largest Tid value and means @head.
const zodb::Tid TidHead = -1ULL;


// v mimics %v for T to be used in printf & friends.
//
// NOTE returned char* pointer is guaranteed to stay valid only till end of
// current expression. For example
//
//	printf("hello %s", v(obj))
//
// is valid, while
//
//	x = v(obj);
//	use(x);
//
// is not valid.
#define v(obj)	(wcfs::v_(obj).c_str())
template<typename T> string v_(T obj) {
    return obj.String();
}

template<> inline string v_(string s) { return s; } // XXX -> const string& ?
template<> string v_(error);
template<> string v_(zodb::Tid);
template<> string v_(zodb::Oid);

}   // wcfs::

#endif
